#ifndef _H_LIBCX_LOGGER
#define _H_LIBCX_LOGGER

#include <cx/io/print.hpp>
#include <cx/string.hpp>
#include <cx/sync/spinlock.hpp>
#include <cx/format/fmt.hpp>

namespace log {
#define LOG_LEVEL_COUNT 4

    enum LogLevel { LOG_OFF = -1, LOG_ERROR = 0, LOG_WARN, LOG_INFO, LOG_DEBUG };

    class Logger {
    private:
        LogLevel filter_level = LOG_INFO;
        const char* prefixes[LOG_LEVEL_COUNT];

    public:
        void log(LogLevel level, const char* file, const int line, const char* func, const char* str);
        void log(LogLevel level, const char* file, const int line, const char* func, String str);

        void set_log_level(LogLevel level);
        const char* get_prefix(LogLevel level);
        void load_default_prefixes_with_color();
        void load_default_prefixes();
        void load_custom_prefixes(const char* pfx[LOG_LEVEL_COUNT]);

        Logger();
        ~Logger() = default;

    private:
        bool should_display(LogLevel level);
    };

    namespace {
        SpinLock<log::Logger*> LOGGER = SpinLock<log::Logger*>(new log::Logger());
    }

    template <typename... Args>
    void debug(const char* fmt, Args... args, const char* file = __builtin_FILE(), int line = __builtin_LINE(), const char* function = __builtin_FUNCTION()) {
        LOGGER.lock().get()->log(LOG_DEBUG, file, line, function, format(fmt, args...));
    }

    template <typename... Args>
    void info(const char* fmt, Args... args, const char* file = __builtin_FILE(), int line = __builtin_LINE(), const char* function = __builtin_FUNCTION()) {
        LOGGER.lock().get()->log(LOG_INFO, file, line, function, format(fmt, args...));
    }

    template <typename... Args>
    void warn(const char* fmt, Args... args, const char* file = __builtin_FILE(), int line = __builtin_LINE(), const char* function = __builtin_FUNCTION()) {
        LOGGER.lock().get()->log(LOG_WARN, file, line, function, format(fmt, args...));
    }

    template <typename... Args>
    void error(const char* fmt, Args... args, const char* file = __builtin_FILE(), int line = __builtin_LINE(), const char* function = __builtin_FUNCTION()) {
        LOGGER.lock().get()->log(LOG_ERROR, file, line, function, format(fmt, args...));
    }
} // namespace log

#endif // !_H_LIBCX_LOGGER
