#ifndef _H_LIBCX_RESULT
#define _H_LIBCX_RESULT

#include <cx/stl.hpp>

extern "C" {
int puts(const char* s);
};

template <typename E>
struct ResultErrBase {
    bool is_okay = true;
    E err_data;

    void set_err(const E& e) {
        this->err_data = stl::move(e);
        this->is_okay = false;
    }

    void set_err(E&& e) {
        this->err_data = stl::move(e);
        this->is_okay = false;
    }

    ~ResultErrBase() {
        if (!this->is_okay) {
            this->err_data.~E();
        }
    }

    bool is_ok() const {
        return this->is_okay;
    }

    bool is_err() const {
        return !this->is_okay;
    }

    E& unwrap_err() {
        if (is_okay) {
            throw "Unwrapped result with error";
        }
        return &this->err_data;
    }
};

template <typename T, typename E>
class Result : private ResultErrBase<E> {
    T ok_data;

public:
    static Result ok(const T& val) {
        Result r;
        r.ok_data = stl::move(val);
        return r;
    }

    static Result ok(T&& val) {
        Result r;
        r.ok_data = stl::move(val);
        return r;
    }

    using ResultErrBase<E>::set_err;
    using ResultErrBase<E>::is_ok;
    using ResultErrBase<E>::is_err;
    using ResultErrBase<E>::unwrap_err;

    static Result err(const E& e) {
        Result r;
        r.set_err(e);
        return r;
    }

    static Result err(E&& e) {
        Result r;
        r.set_err(stl::move(e));
        return r;
    }

    ~Result() {
        if (this->is_ok()) {
            this->ok_data.~T();
        }
    }

    T& expect(const char* str) {
        if (!this->is_ok()) {
            puts("PANIC: Unwrapped result with error: ");
            puts(str);
            puts("\n");
            __builtin_trap();
        }
        return &this->ok_data;
    }

    T& unwrap() {
        if (!this->is_ok()) {
            puts("PANIC: Unwrapped result with error\n");
            __builtin_trap();
        }
        return &this->ok_data;
    }
};

template <typename E>
class Result<void, E> : private ResultErrBase<E> {
public:
    static Result ok() {
        return Result{};
    }

    using ResultErrBase<E>::set_err;
    using ResultErrBase<E>::is_ok;
    using ResultErrBase<E>::is_err;
    using ResultErrBase<E>::unwrap_err;

    static Result err(const E& e) {
        Result r;
        r.set_err(e);
        return r;
    }

    static Result err(E&& e) {
        Result r;
        r.set_err(stl::move(e));
        return r;
    }

    void unwrap() {
        if (!this->is_ok()) {
            throw "Unwrapped result with error";
        }
    }
};

#endif // _H_LIBCX_RESULT