#ifndef _H_LIBCX_SYNC_MUTEX
#define _H_LIBCX_SYNC_MUTEX
#include <cx/config/platform.hpp>

#ifdef PLATFORM_LINUX
#include <linux/futex.h>
#include <sys/syscall.h>
#endif
#include <unistd.h>

template <typename T>
struct Mutex {
private:
    int locked = 0;
    T value;

    Mutex() = default;

    void acquire() {
#ifdef PLATFORM_LINUX
        int expected = 0;
        if (!__atomic_compare_exchange_n(&this->locked, &expected, 1, false, __ATOMIC_ACQUIRE, __ATOMIC_RELAXED)) {
            // someone else already locked it, wait
            while (__atomic_exchange_n(&this->locked, 2, __ATOMIC_ACQUIRE) != 0) {
                syscall(SYS_futex, &this->locked, FUTEX_WAIT, 2, nullptr, nullptr, 0);
            }
        }
#endif
    }

    void release() {
#ifdef PLATFORM_LINUX
        if (__atomic_exchange_n(&this->locked, 0, __ATOMIC_RELEASE) == 2) {
            syscall(SYS_futex, &this->locked, FUTEX_WAKE, 1, nullptr, nullptr, 0);
        }
#endif
    }

    struct MutexGuard {
    private:
        Mutex<T>* parent;

    public:
        MutexGuard(Mutex<T>* p) : parent(p) {
            parent->acquire();
        }

        ~MutexGuard() {
            parent->release();
        }

        T& get() {
            return parent->value;
        }

        T* operator->() {
            return &parent->value;
        }

        T& operator*() {
            return parent->value;
        }

        MutexGuard(const MutexGuard&) = delete;
        MutexGuard& operator=(const MutexGuard&) = delete;
    };

public:
    template <typename... Args>
    explicit Mutex(Args&&... args) : value((Args&&)args...) {}

    MutexGuard lock() {
        return MutexGuard(this);
    }
};

#endif // !_H_LIBCX_SYNC_MUTEX
